/**
 *
 * $Id: misc.c,v 1.4 1996/11/08 04:46:31 miers Exp $
 *
 * Copyright (C) 1995 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

/*
 * This file contains routines that just don't belong anywhere else
 */

#include <LTconfig.h>
#include <XmI/XmI.h>

#include <Xm/XmP.h>
#if XtSpecificationRelease < 6
#include <X11/IntrinsicP.h>
#endif

#ifndef HAVE_STRCASECMP
int
strcasecmp(s1, s2)
    char *s1, *s2;
{
    int c1, c2;

    while (*s1 && *s2) {
	c1 = tolower(*s1);
	c2 = tolower(*s2);
	if (c1 != c2)
	    return (c1 - c2);
	s1++;
	s2++;
    }
    return (int) (*s1 - *s2);
}
#endif

#ifndef HAVE_STRNCASECMP
int
strncasecmp(s1, s2, count)
    char *s1, *s2;
    int count;
{
    int c1, c2;
    int i = 0;

    while (*s1 && *s2 && i < count) {
	c1 = tolower(*s1);
	c2 = tolower(*s2);
	if (c1 != c2)
	    return (c1 - c2);
	s1++;
	s2++;
	i++;
    }
    return (int) (*s1 - *s2);
}
#endif

#ifndef HAVE_STRSTR
char *
strstr(s1, s2)
    char *s1;
    char *s2;
{
    int i;
    char *p1;
    char *p2;
    char *s = s1;

    for (p2 = s2, i = 0; *s; p2 = s2, i++, s++)
    {
	for (p1 = s; *p1 && *p2 && *p1 == *p2; p1++, p2++)
	    ;
	if (!*p2)
	    break;
    }
    if (!*p2)
	return s1 + i;

    return 0;
}
#endif

void
XmUpdateDisplay(Widget w)
{
    XEvent  ev;
    Display * dsp;
 
    dsp = XtDisplay(w);
    
    /* First process all available events ... */
    while ( XCheckMaskEvent(dsp, ExposureMask, &ev))
	XtDispatchEvent(&ev);
    
    /* Flush all buffers */
    XSync(dsp,False);
    
    /* Process remaining events ... */
    while ( XCheckMaskEvent(dsp, ExposureMask, &ev))
	XtDispatchEvent(&ev);
}

#define even_stipple_width 8
#define even_stipple_height 8
static unsigned char even_stipple_bits[] = {
   0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55 };
#define odd_stipple_width 8
#define odd_stipple_height 8
static unsigned char odd_stipple_bits[] = {
   0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa };

void
_XmInstallStippleImages(Widget w) {
    XImage *even_stipple_image;
    XImage *odd_stipple_image;

    _XmCreateImage(even_stipple_image, XtDisplay(w), (char *)even_stipple_bits,
		   even_stipple_width, even_stipple_height, LSBFirst);
    XmInstallImage(even_stipple_image, XmEVEN_STIPPLE_IMAGE);

    _XmCreateImage(odd_stipple_image, XtDisplay(w), (char *)odd_stipple_bits,
		   odd_stipple_width, odd_stipple_height, LSBFirst);
    XmInstallImage(odd_stipple_image, XmODD_STIPPLE_IMAGE);
}

/*
 * The missing commas are intentional.  Let the compiler
 * concatenate the strings
 */
_XmConst char _XmStrings[] =
/* 0     */ "\0"
/* 1     */ "Accelerator\0"
/* 13    */ "AcceleratorText\0"
/* 29    */ "AdjustLast\0"
/* 40    */ "AdjustMargin\0"
/* 53    */ "Alignment\0"
/* 63    */ "AllowOverlap\0"
/* 76    */ "AnimationMask\0"
/* 90    */ "AnimationPixmap\0"
/* 106   */ "AnimationPixmapDepth\0"
/* 127   */ "AnimationStyle\0"
/* 142   */ "ApplyLabelString\0"
/* 159   */ "ArmCallback\0"
/* 171   */ "ArmColor\0"
/* 180   */ "ArmPixmap\0"
/* 190   */ "ArrowDirection\0"
/* 205   */ "Attachment\0"
/* 216   */ "AudibleWarning\0"
/* 231   */ "AutoShowCursorPosition\0"
/* 254   */ "AutoUnmanage\0"
/* 267   */ "AutomaticSelection\0"
/* 286   */ "Availability\0"
/* 299   */ "BackgroundPixmap\0"
/* 316   */ "BlendModel\0"
/* 327   */ "BlinkRate\0"
/* 337   */ "BottomShadowColor\0"
/* 355   */ "BottomShadowPixmap\0"
/* 374   */ "ButtonAcceleratorText\0"
/* 396   */ "ButtonAccelerators\0"
/* 415   */ "ButtonCount\0"
/* 427   */ "ButtonFontList\0"
/* 442   */ "ButtonMnemonicCharSets\0"
/* 465   */ "ButtonMnemonics\0"
/* 481   */ "ButtonSet\0"
/* 491   */ "ButtonType\0"
/* 502   */ "Buttons\0"
/* 510   */ "CancelLabelString\0"
/* 528   */ "ChildHorizontalAlignment\0"
/* 553   */ "ChildHorizontalSpacing\0"
/* 576   */ "ChildPlacement\0"
/* 591   */ "ChildType\0"
/* 601   */ "ChildVerticalAlignment\0"
/* 624   */ "Children\0"
/* 633   */ "ClientData\0"
/* 644   */ "ClipWindow\0"
/* 655   */ "Columns\0"
/* 663   */ "CommandWindow\0"
/* 677   */ "CommandWindowLocation\0"
/* 699   */ "ConvertProc\0"
/* 711   */ "CursorBackground\0"
/* 728   */ "CursorForeground\0"
/* 745   */ "CursorPosition\0"
/* 760   */ "CursorPositionVisible\0"
/* 782   */ "DarkThreshold\0"
/* 796   */ "DecimalPoints\0"
/* 810   */ "DefaultButtonShadowThickness\0"
/* 839   */ "DefaultButtonType\0"
/* 857   */ "DefaultCopyCursorIcon\0"
/* 879   */ "DefaultFontList\0"
/* 895   */ "DefaultInvalidCursorIcon\0"
/* 920   */ "DefaultLinkCursorIcon\0"
/* 942   */ "DefaultMoveCursorIcon\0"
/* 964   */ "DefaultNoneCursorIcon\0"
/* 986   */ "DefaultPosition\0"
/* 1002  */ "DefaultSourceCursorIcon\0"
/* 1026  */ "DefaultValidCursorIcon\0"
/* 1049  */ "DeleteResponse\0"
/* 1064  */ "DesktopParent\0"
/* 1078  */ "DialogStyle\0"
/* 1090  */ "DialogTitle\0"
/* 1102  */ "DialogType\0"
/* 1113  */ "DirListItemCount\0"
/* 1130  */ "DirListItems\0"
/* 1143  */ "DirListLabelString\0"
/* 1162  */ "DirMask\0"
/* 1170  */ "DirSearchProc\0"
/* 1184  */ "DirSpec\0"
/* 1192  */ "Directory\0"
/* 1202  */ "DirectoryValid\0"
/* 1217  */ "DisarmCallback\0"
/* 1232  */ "DoubleClickInterval\0"
/* 1252  */ "DragContextClass\0"
/* 1269  */ "DragDropFinishCallback\0"
/* 1292  */ "DragIconClass\0"
/* 1306  */ "DragInitiatorProtocolStyle\0"
/* 1333  */ "DragMotionCallback\0"
/* 1352  */ "DragOperations\0"
/* 1367  */ "DragOverMode\0"
/* 1380  */ "DragProc\0"
/* 1389  */ "DragReceiverProtocolStyle\0"
/* 1415  */ "DropProc\0"
/* 1424  */ "DropRectangles\0"
/* 1439  */ "DropSiteActivity\0"
/* 1456  */ "DropSiteEnterCallback\0"
/* 1478  */ "DropSiteLeaveCallback\0"
/* 1500  */ "DropSiteManagerClass\0"
/* 1521  */ "DropSiteOperations\0"
/* 1540  */ "DropSiteType\0"
/* 1553  */ "DropStartCallback\0"
/* 1571  */ "DropTransferClass\0"
/* 1589  */ "DropTransfers\0"
/* 1603  */ "Editable\0"
/* 1612  */ "EntryBorder\0"
/* 1624  */ "EntryClass\0"
/* 1635  */ "ExportTargets\0"
/* 1649  */ "ExposeCallback\0"
/* 1664  */ "ExtensionType\0"
/* 1678  */ "FileListItemCount\0"
/* 1696  */ "FileListItems\0"
/* 1710  */ "FileListLabelString\0"
/* 1730  */ "FileSearchProc\0"
/* 1745  */ "FileTypeMask\0"
/* 1758  */ "FillOnArm\0"
/* 1768  */ "FillOnSelect\0"
/* 1781  */ "FilterLabelString\0"
/* 1799  */ "FontList\0"
/* 1808  */ "ForegroundThreshold\0"
/* 1828  */ "HelpLabelString\0"
/* 1844  */ "HighlightColor\0"
/* 1859  */ "HighlightOnEnter\0"
/* 1876  */ "HighlightPixmap\0"
/* 1892  */ "HighlightThickness\0"
/* 1911  */ "HorizontalFontUnit\0"
/* 1930  */ "HorizontalScrollBar\0"
/* 1950  */ "Hot\0"
/* 1954  */ "ICCHandle\0"
/* 1964  */ "ImportTargets\0"
/* 1978  */ "Increment\0"
/* 1988  */ "Incremental\0"
/* 2000  */ "IndicatorOn\0"
/* 2012  */ "IndicatorSize\0"
/* 2026  */ "IndicatorType\0"
/* 2040  */ "InitialDelay\0"
/* 2053  */ "InitialFocus\0"
/* 2066  */ "InputCreate\0"
/* 2078  */ "InputMethod\0"
/* 2090  */ "InvalidCursorForeground\0"
/* 2114  */ "IsAligned\0"
/* 2124  */ "IsHomogeneous\0"
/* 2138  */ "ItemCount\0"
/* 2148  */ "Items\0"
/* 2154  */ "KeyboardFocusPolicy\0"
/* 2174  */ "LabelFontList\0"
/* 2188  */ "LabelInsensitivePixmap\0"
/* 2211  */ "LabelPixmap\0"
/* 2223  */ "LabelString\0"
/* 2235  */ "LabelType\0"
/* 2245  */ "LightThreshold\0"
/* 2260  */ "ListLabelString\0"
/* 2276  */ "ListMarginHeight\0"
/* 2293  */ "ListMarginWidth\0"
/* 2309  */ "ListSizePolicy\0"
/* 2324  */ "ListSpacing\0"
/* 2336  */ "ListUpdated\0"
/* 2348  */ "LogicalParent\0"
/* 2362  */ "MainWindowMarginHeight\0"
/* 2385  */ "MainWindowMarginWidth\0"
/* 2407  */ "MappingDelay\0"
/* 2420  */ "MarginBottom\0"
/* 2433  */ "MarginHeight\0"
/* 2446  */ "MarginLeft\0"
/* 2457  */ "MarginRight\0"
/* 2469  */ "MarginTop\0"
/* 2479  */ "MarginWidth\0"
/* 2491  */ "Mask\0"
/* 2496  */ "MaxItems\0"
/* 2505  */ "MaxLength\0"
/* 2515  */ "MaxValue\0"
/* 2524  */ "Maximum\0"
/* 2532  */ "MenuBar\0"
/* 2540  */ "MenuPost\0"
/* 2549  */ "MenuWidget\0"
/* 2560  */ "MessageProc\0"
/* 2572  */ "MessageWindow\0"
/* 2586  */ "MinimizeButtons\0"
/* 2602  */ "Minimum\0"
/* 2610  */ "Mnemonic\0"
/* 2619  */ "MnemonicCharSet\0"
/* 2635  */ "MoveOpaque\0"
/* 2646  */ "MultiClick\0"
/* 2657  */ "MustMatch\0"
/* 2667  */ "MwmDecorations\0"
/* 2682  */ "MwmFunctions\0"
/* 2695  */ "MwmInputMode\0"
/* 2708  */ "MwmMenu\0"
/* 2716  */ "MwmMessages\0"
/* 2728  */ "NavigationType\0"
/* 2743  */ "NeedsMotion\0"
/* 2755  */ "NoMatchString\0"
/* 2769  */ "NoResize\0"
/* 2778  */ "NoneCursorForeground\0"
/* 2799  */ "NotifyProc\0"
/* 2810  */ "NumChildren\0"
/* 2822  */ "NumColumns\0"
/* 2833  */ "NumDropRectangles\0"
/* 2851  */ "NumDropTransfers\0"
/* 2868  */ "NumExportTargets\0"
/* 2885  */ "NumImportTargets\0"
/* 2902  */ "Offset\0"
/* 2909  */ "OkLabelString\0"
/* 2923  */ "OperationChangedCallback\0"
/* 2948  */ "OperationCursorIcon\0"
/* 2968  */ "OptionLabel\0"
/* 2980  */ "OptionMnemonic\0"
/* 2995  */ "OutputCreate\0"
/* 3008  */ "Packing\0"
/* 3016  */ "PageIncrement\0"
/* 3030  */ "PaneMaximum\0"
/* 3042  */ "PaneMinimum\0"
/* 3054  */ "Pattern\0"
/* 3062  */ "PendingDelete\0"
/* 3076  */ "PopupEnabled\0"
/* 3089  */ "PositionIndex\0"
/* 3103  */ "PostFromButton\0"
/* 3118  */ "PostFromCount\0"
/* 3132  */ "PostFromList\0"
/* 3145  */ "PreeditType\0"
/* 3157  */ "ProcessingDirection\0"
/* 3177  */ "PromptString\0"
/* 3190  */ "ProtocolCallback\0"
/* 3207  */ "PushButtonEnabled\0"
/* 3225  */ "QualifySearchDataProc\0"
/* 3247  */ "RadioAlwaysOne\0"
/* 3262  */ "RadioBehavior\0"
/* 3276  */ "RecomputeSize\0"
/* 3290  */ "Rectangles\0"
/* 3301  */ "RepeatDelay\0"
/* 3313  */ "ResizeCallback\0"
/* 3328  */ "ResizeHeight\0"
/* 3341  */ "ResizePolicy\0"
/* 3354  */ "ResizeWidth\0"
/* 3366  */ "RowColumnType\0"
/* 3380  */ "Rows\0"
/* 3385  */ "RubberPositioning\0"
/* 3403  */ "SashHeight\0"
/* 3414  */ "SashIndent\0"
/* 3425  */ "SashWidth\0"
/* 3435  */ "ScaleHeight\0"
/* 3447  */ "ScaleMultiple\0"
/* 3461  */ "ScaleWidth\0"
/* 3472  */ "Scroll\0"
/* 3479  */ "ScrollBarDisplayPolicy\0"
/* 3502  */ "ScrollBarPlacement\0"
/* 3521  */ "ScrollSide\0"
/* 3532  */ "ScrolledWindowMarginHeight\0"
/* 3559  */ "ScrolledWindowMarginWidth\0"
/* 3585  */ "ScrollingPolicy\0"
/* 3601  */ "SelectColor\0"
/* 3613  */ "SelectInsensitivePixmap\0"
/* 3637  */ "SelectPixmap\0"
/* 3650  */ "SelectThreshold\0"
/* 3666  */ "SelectedItemCount\0"
/* 3684  */ "SelectedItems\0"
/* 3698  */ "SelectionArrayCount\0"
/* 3718  */ "SelectionLabelString\0"
/* 3739  */ "SelectionPolicy\0"
/* 3755  */ "SeparatorOn\0"
/* 3767  */ "SeparatorType\0"
/* 3781  */ "Set\0"
/* 3785  */ "ShadowThickness\0"
/* 3801  */ "ShadowType\0"
/* 3812  */ "ShellUnitType\0"
/* 3826  */ "ShowArrows\0"
/* 3837  */ "ShowAsDefault\0"
/* 3851  */ "ShowSeparator\0"
/* 3865  */ "ShowValue\0"
/* 3875  */ "SimpleCheckBox\0"
/* 3890  */ "SimpleMenuBar\0"
/* 3904  */ "SimpleOptionMenu\0"
/* 3921  */ "SimplePopupMenu\0"
/* 3937  */ "SimplePulldownMenu\0"
/* 3956  */ "SimpleRadioBox\0"
/* 3971  */ "SizePolicy\0"
/* 3982  */ "SliderSize\0"
/* 3993  */ "Source\0"
/* 4000  */ "SourceCursorIcon\0"
/* 4017  */ "SourceIsExternal\0"
/* 4034  */ "SourcePixmapIcon\0"
/* 4051  */ "SourceWidget\0"
/* 4064  */ "SourceWindow\0"
/* 4077  */ "Spacing\0"
/* 4085  */ "StartTime\0"
/* 4095  */ "StateCursorIcon\0"
/* 4111  */ "StringDirection\0"
/* 4127  */ "TearOffModel\0"
/* 4140  */ "TextFontList\0"
/* 4153  */ "TextString\0"
/* 4164  */ "TextValue\0"
/* 4174  */ "TitleString\0"
/* 4186  */ "TopCharacter\0"
/* 4199  */ "TopItemPosition\0"
/* 4215  */ "TopLevelEnterCallback\0"
/* 4237  */ "TopLevelLeaveCallback\0"
/* 4259  */ "TopShadowColor\0"
/* 4274  */ "TopShadowPixmap\0"
/* 4290  */ "TransferProc\0"
/* 4303  */ "TransferStatus\0"
/* 4318  */ "TraversalOn\0"
/* 4330  */ "TraversalType\0"
/* 4344  */ "TreeUpdateProc\0"
/* 4359  */ "TroughColor\0"
/* 4371  */ "UnitType\0"
/* 4380  */ "UnpostBehavior\0"
/* 4395  */ "UnselectPixmap\0"
/* 4410  */ "UpdateSliderSize\0"
/* 4427  */ "UseAsyncGeometry\0"
/* 4444  */ "UserData\0"
/* 4453  */ "ValidCursorForeground\0"
/* 4475  */ "ValueChangedCallback\0"
/* 4496  */ "ValueWcs\0"
/* 4505  */ "VerifyBell\0"
/* 4516  */ "VerticalAlignment\0"
/* 4534  */ "VerticalFontUnit\0"
/* 4551  */ "VerticalScrollBar\0"
/* 4569  */ "VisibleItemCount\0"
/* 4586  */ "VisibleWhenOff\0"
/* 4601  */ "VisualPolicy\0"
/* 4614  */ "WhichButton\0"
/* 4626  */ "WordWrap\0"
/* 4635  */ "WorkWindow\0"
/* 4646  */ "XmString\0"
/* 4655  */ "accelerator\0"
/* 4667  */ "acceleratorText\0"
/* 4683  */ "activateCallback\0"
/* 4700  */ "adjustLast\0"
/* 4711  */ "adjustMargin\0"
/* 4724  */ "alignment\0"
/* 4734  */ "allowOverlap\0"
/* 4747  */ "allowResize\0"
/* 4759  */ "animationMask\0"
/* 4773  */ "animationPixmap\0"
/* 4789  */ "animationPixmapDepth\0"
/* 4810  */ "animationStyle\0"
/* 4825  */ "applyCallback\0"
/* 4839  */ "applyLabelString\0"
/* 4856  */ "armCallback\0"
/* 4868  */ "armColor\0"
/* 4877  */ "armPixmap\0"
/* 4887  */ "arrowDirection\0"
/* 4902  */ "attachment\0"
/* 4913  */ "audibleWarning\0"
/* 4928  */ "autoShowCursorPosition\0"
/* 4951  */ "autoUnmanage\0"
/* 4964  */ "automaticSelection\0"
/* 4983  */ "availability\0"
/* 4996  */ "blendModel\0"
/* 5007  */ "blinkRate\0"
/* 5017  */ "bottomAttachment\0"
/* 5034  */ "bottomOffset\0"
/* 5047  */ "bottomPosition\0"
/* 5062  */ "bottomShadowColor\0"
/* 5080  */ "bottomShadowPixmap\0"
/* 5099  */ "bottomWidget\0"
/* 5112  */ "browseSelectionCallback\0"
/* 5136  */ "buttonAcceleratorText\0"
/* 5158  */ "buttonAccelerators\0"
/* 5177  */ "buttonCount\0"
/* 5189  */ "buttonFontList\0"
/* 5204  */ "buttonMnemonicCharSets\0"
/* 5227  */ "buttonMnemonics\0"
/* 5243  */ "buttonSet\0"
/* 5253  */ "buttonType\0"
/* 5264  */ "buttons\0"
/* 5272  */ "cancelButton\0"
/* 5285  */ "cancelCallback\0"
/* 5300  */ "cancelLabelString\0"
/* 5318  */ "cascadePixmap\0"
/* 5332  */ "cascadingCallback\0"
/* 5350  */ "childHorizontalAlignment\0"
/* 5375  */ "childHorizontalSpacing\0"
/* 5398  */ "childPlacement\0"
/* 5413  */ "childPosition\0"
/* 5427  */ "childType\0"
/* 5437  */ "childVerticalAlignment\0"
/* 5460  */ "clientData\0"
/* 5471  */ "clipWindow\0"
/* 5482  */ "columns\0"
/* 5490  */ "command\0"
/* 5498  */ "commandChangedCallback\0"
/* 5521  */ "commandEnteredCallback\0"
/* 5544  */ "commandWindow\0"
/* 5558  */ "commandWindowLocation\0"
/* 5580  */ "convertProc\0"
/* 5592  */ "cursorBackground\0"
/* 5609  */ "cursorForeground\0"
/* 5626  */ "cursorPosition\0"
/* 5641  */ "cursorPositionVisible\0"
/* 5663  */ "darkThreshold\0"
/* 5677  */ "decimalPoints\0"
/* 5691  */ "decrementCallback\0"
/* 5709  */ "defaultActionCallback\0"
/* 5731  */ "defaultButton\0"
/* 5745  */ "defaultButtonShadowThickness\0"
/* 5774  */ "defaultButtonType\0"
/* 5792  */ "defaultCopyCursorIcon\0"
/* 5814  */ "defaultFontList\0"
/* 5830  */ "defaultInvalidCursorIcon\0"
/* 5855  */ "defaultLinkCursorIcon\0"
/* 5877  */ "defaultMoveCursorIcon\0"
/* 5899  */ "defaultNoneCursorIcon\0"
/* 5921  */ "defaultPosition\0"
/* 5937  */ "defaultSourceCursorIcon\0"
/* 5961  */ "defaultValidCursorIcon\0"
/* 5984  */ "deleteResponse\0"
/* 5999  */ "desktopParent\0"
/* 6013  */ "dialogStyle\0"
/* 6025  */ "dialogTitle\0"
/* 6037  */ "dialogType\0"
/* 6048  */ "dirListItemCount\0"
/* 6065  */ "dirListItems\0"
/* 6078  */ "dirListLabelString\0"
/* 6097  */ "dirMask\0"
/* 6105  */ "dirSearchProc\0"
/* 6119  */ "dirSpec\0"
/* 6127  */ "directory\0"
/* 6137  */ "directoryValid\0"
/* 6152  */ "disarmCallback\0"
/* 6167  */ "doubleClickInterval\0"
/* 6187  */ "dragCallback\0"
/* 6200  */ "dragContextClass\0"
/* 6217  */ "dragDropFinishCallback\0"
/* 6240  */ "dragIconClass\0"
/* 6254  */ "dragInitiatorProtocolStyle\0"
/* 6281  */ "dragMotionCallback\0"
/* 6300  */ "dragOperations\0"
/* 6315  */ "dragOverMode\0"
/* 6328  */ "dragProc\0"
/* 6337  */ "dragReceiverProtocolStyle\0"
/* 6363  */ "dropFinishCallback\0"
/* 6382  */ "dropProc\0"
/* 6391  */ "dropRectangles\0"
/* 6406  */ "dropSiteActivity\0"
/* 6423  */ "dropSiteEnterCallback\0"
/* 6445  */ "dropSiteLeaveCallback\0"
/* 6467  */ "dropSiteManagerClass\0"
/* 6488  */ "dropSiteOperations\0"
/* 6507  */ "dropSiteType\0"
/* 6520  */ "dropStartCallback\0"
/* 6538  */ "dropTransferClass\0"
/* 6556  */ "dropTransfers\0"
/* 6570  */ "editMode\0"
/* 6579  */ "editable\0"
/* 6588  */ "entryAlignment\0"
/* 6603  */ "entryBorder\0"
/* 6615  */ "entryCallback\0"
/* 6629  */ "entryClass\0"
/* 6640  */ "entryVerticalAlignment\0"
/* 6663  */ "exportTargets\0"
/* 6677  */ "exposeCallback\0"
/* 6692  */ "extendedSelectionCallback\0"
/* 6718  */ "extensionType\0"
/* 6732  */ "fileListItemCount\0"
/* 6750  */ "fileListItems\0"
/* 6764  */ "fileListLabelString\0"
/* 6784  */ "fileSearchProc\0"
/* 6799  */ "fileTypeMask\0"
/* 6812  */ "fillOnArm\0"
/* 6822  */ "fillOnSelect\0"
/* 6835  */ "filterLabelString\0"
/* 6853  */ "focusCallback\0"
/* 6867  */ "focusMovedCallback\0"
/* 6886  */ "focusPolicyChanged\0"
/* 6905  */ "fontList\0"
/* 6914  */ "foregroundThreshold\0"
/* 6934  */ "fractionBase\0"
/* 6947  */ "gainPrimaryCallback\0"
/* 6967  */ "helpCallback\0"
/* 6980  */ "helpLabelString\0"
/* 6996  */ "highlightColor\0"
/* 7011  */ "highlightOnEnter\0"
/* 7028  */ "highlightPixmap\0"
/* 7044  */ "highlightThickness\0"
/* 7063  */ "historyItemCount\0"
/* 7080  */ "historyItems\0"
/* 7093  */ "historyMaxItems\0"
/* 7109  */ "historyVisibleItemCount\0"
/* 7133  */ "horizontalFontUnit\0"
/* 7152  */ "horizontalScrollBar\0"
/* 7172  */ "horizontalSpacing\0"
/* 7190  */ "hotX\0"
/* 7195  */ "hotY\0"
/* 7200  */ "iccHandle\0"
/* 7210  */ "importTargets\0"
/* 7224  */ "increment\0"
/* 7234  */ "incrementCallback\0"
/* 7252  */ "incremental\0"
/* 7264  */ "indicatorOn\0"
/* 7276  */ "indicatorSize\0"
/* 7290  */ "indicatorType\0"
/* 7304  */ "initialDelay\0"
/* 7317  */ "initialFocus\0"
/* 7330  */ "inputCallback\0"
/* 7344  */ "inputCreate\0"
/* 7356  */ "inputMethod\0"
/* 7368  */ "invalidCursorForeground\0"
/* 7392  */ "isAligned\0"
/* 7402  */ "isHomogeneous\0"
/* 7416  */ "itemCount\0"
/* 7426  */ "items\0"
/* 7432  */ "keyboardFocusPolicy\0"
/* 7452  */ "labelFontList\0"
/* 7466  */ "labelInsensitivePixmap\0"
/* 7489  */ "labelPixmap\0"
/* 7501  */ "labelString\0"
/* 7513  */ "labelType\0"
/* 7523  */ "leftAttachment\0"
/* 7538  */ "leftOffset\0"
/* 7549  */ "leftPosition\0"
/* 7562  */ "leftWidget\0"
/* 7573  */ "lightThreshold\0"
/* 7588  */ "lineSpace\0"
/* 7598  */ "listItemCount\0"
/* 7612  */ "listItems\0"
/* 7622  */ "listLabelString\0"
/* 7638  */ "listMarginHeight\0"
/* 7655  */ "listMarginWidth\0"
/* 7671  */ "listSizePolicy\0"
/* 7686  */ "listSpacing\0"
/* 7698  */ "listUpdated\0"
/* 7710  */ "listVisibleItemCount\0"
/* 7731  */ "logicalParent\0"
/* 7745  */ "losePrimaryCallback\0"
/* 7765  */ "losingFocusCallback\0"
/* 7785  */ "mainWindowMarginHeight\0"
/* 7808  */ "mainWindowMarginWidth\0"
/* 7830  */ "mapCallback\0"
/* 7842  */ "mappingDelay\0"
/* 7855  */ "margin\0"
/* 7862  */ "marginBottom\0"
/* 7875  */ "marginHeight\0"
/* 7888  */ "marginLeft\0"
/* 7899  */ "marginRight\0"
/* 7911  */ "marginTop\0"
/* 7921  */ "marginWidth\0"
/* 7933  */ "mask\0"
/* 7938  */ "maxLength\0"
/* 7948  */ "maximum\0"
/* 7956  */ "menuAccelerator\0"
/* 7972  */ "menuBar\0"
/* 7980  */ "menuCursor\0"
/* 7991  */ "menuHelpWidget\0"
/* 8006  */ "menuHistory\0"
/* 8018  */ "menuPost\0"
/* 8027  */ "messageAlignment\0"
/* 8044  */ "messageProc\0"
/* 8056  */ "messageString\0"
/* 8070  */ "messageWindow\0"
/* 8084  */ "minimizeButtons\0"
/* 8100  */ "minimum\0"
/* 8108  */ "mnemonic\0"
/* 8117  */ "mnemonicCharSet\0"
/* 8133  */ "modifyVerifyCallback\0"
/* 8154  */ "modifyVerifyCallbackWcs\0"
/* 8178  */ "motionVerifyCallback\0"
/* 8199  */ "moveOpaque\0"
/* 8210  */ "multiClick\0"
/* 8221  */ "multipleSelectionCallback\0"
/* 8247  */ "mustMatch\0"
/* 8257  */ "mwmDecorations\0"
/* 8272  */ "mwmFunctions\0"
/* 8285  */ "mwmInputMode\0"
/* 8298  */ "mwmMenu\0"
/* 8306  */ "mwmMessages\0"
/* 8318  */ "navigationType\0"
/* 8333  */ "needsMotion\0"
/* 8345  */ "noMatchCallback\0"
/* 8361  */ "noMatchString\0"
/* 8375  */ "noResize\0"
/* 8384  */ "noneCursorForeground\0"
/* 8405  */ "notifyProc\0"
/* 8416  */ "numColumns\0"
/* 8427  */ "numDropRectangles\0"
/* 8445  */ "numDropTransfers\0"
/* 8462  */ "numExportTargets\0"
/* 8479  */ "numImportTargets\0"
/* 8496  */ "numRectangles\0"
/* 8510  */ "offsetX\0"
/* 8518  */ "offsetY\0"
/* 8526  */ "okCallback\0"
/* 8537  */ "okLabelString\0"
/* 8551  */ "operationChangedCallback\0"
/* 8576  */ "operationCursorIcon\0"
/* 8596  */ "optionLabel\0"
/* 8608  */ "optionMnemonic\0"
/* 8623  */ "outputCreate\0"
/* 8636  */ "packing\0"
/* 8644  */ "pageDecrementCallback\0"
/* 8666  */ "pageIncrement\0"
/* 8680  */ "pageIncrementCallback\0"
/* 8702  */ "paneMaximum\0"
/* 8714  */ "paneMinimum\0"
/* 8726  */ "pattern\0"
/* 8734  */ "pendingDelete\0"
/* 8748  */ "popupEnabled\0"
/* 8761  */ "positionIndex\0"
/* 8775  */ "postFromButton\0"
/* 8790  */ "postFromCount\0"
/* 8804  */ "postFromList\0"
/* 8817  */ "preeditType\0"
/* 8829  */ "processingDirection\0"
/* 8849  */ "promptString\0"
/* 8862  */ "protocolCallback\0"
/* 8879  */ "pushButtonEnabled\0"
/* 8897  */ "qualifySearchDataProc\0"
/* 8919  */ "radioAlwaysOne\0"
/* 8934  */ "radioBehavior\0"
/* 8948  */ "realizeCallback\0"
/* 8964  */ "recomputeSize\0"
/* 8978  */ "rectangles\0"
/* 8989  */ "refigureMode\0"
/* 9002  */ "repeatDelay\0"
/* 9014  */ "resizable\0"
/* 9024  */ "resizeCallback\0"
/* 9039  */ "resizeHeight\0"
/* 9052  */ "resizePolicy\0"
/* 9065  */ "resizeWidth\0"
/* 9077  */ "rightAttachment\0"
/* 9093  */ "rightOffset\0"
/* 9105  */ "rightPosition\0"
/* 9119  */ "rightWidget\0"
/* 9131  */ "rowColumnType\0"
/* 9145  */ "rows\0"
/* 9150  */ "rubberPositioning\0"
/* 9168  */ "sashHeight\0"
/* 9179  */ "sashIndent\0"
/* 9190  */ "sashShadowThickness\0"
/* 9210  */ "sashWidth\0"
/* 9220  */ "scaleHeight\0"
/* 9232  */ "scaleMultiple\0"
/* 9246  */ "scaleWidth\0"
/* 9257  */ "scrollBarDisplayPolicy\0"
/* 9280  */ "scrollBarPlacement\0"
/* 9299  */ "scrollHorizontal\0"
/* 9316  */ "scrollLeftSide\0"
/* 9331  */ "scrollTopSide\0"
/* 9345  */ "scrollVertical\0"
/* 9360  */ "scrolledWindowMarginHeight\0"
/* 9387  */ "scrolledWindowMarginWidth\0"
/* 9413  */ "scrollingPolicy\0"
/* 9429  */ "selectColor\0"
/* 9441  */ "selectInsensitivePixmap\0"
/* 9465  */ "selectPixmap\0"
/* 9478  */ "selectThreshold\0"
/* 9494  */ "selectedItemCount\0"
/* 9512  */ "selectedItems\0"
/* 9526  */ "selectionArrayCount\0"
/* 9546  */ "selectionLabelString\0"
/* 9567  */ "selectionPolicy\0"
/* 9583  */ "separatorOn\0"
/* 9595  */ "separatorType\0"
/* 9609  */ "set\0"
/* 9613  */ "shadow\0"
/* 9620  */ "shadowThickness\0"
/* 9636  */ "shadowType\0"
/* 9647  */ "shellUnitType\0"
/* 9661  */ "showArrows\0"
/* 9672  */ "showAsDefault\0"
/* 9686  */ "showSeparator\0"
/* 9700  */ "showValue\0"
/* 9710  */ "simpleCallback\0"
/* 9725  */ "singleSelectionCallback\0"
/* 9749  */ "sizePolicy\0"
/* 9760  */ "skipAdjust\0"
/* 9771  */ "sliderSize\0"
/* 9782  */ "source\0"
/* 9789  */ "sourceCursorIcon\0"
/* 9806  */ "sourceIsExternal\0"
/* 9823  */ "sourcePixmapIcon\0"
/* 9840  */ "sourceWidget\0"
/* 9853  */ "sourceWindow\0"
/* 9866  */ "spacing\0"
/* 9874  */ "spotLocation\0"
/* 9887  */ "startTime\0"
/* 9897  */ "stateCursorIcon\0"
/* 9913  */ "stringDirection\0"
/* 9929  */ "subMenuId\0"
/* 9939  */ "symbolPixmap\0"
/* 9952  */ "tearOffMenuActivateCallback\0"
/* 9980  */ "tearOffMenuDeactivateCallback\0"
/* 10010 */ "tearOffModel\0"
/* 10023 */ "textAccelerators\0"
/* 10040 */ "textColumns\0"
/* 10052 */ "textFontList\0"
/* 10065 */ "textString\0"
/* 10076 */ "textTranslations\0"
/* 10093 */ "textValue\0"
/* 10103 */ "titleString\0"
/* 10115 */ "toBottomCallback\0"
/* 10132 */ "toPositionCallback\0"
/* 10151 */ "toTopCallback\0"
/* 10165 */ "topAttachment\0"
/* 10179 */ "topCharacter\0"
/* 10192 */ "topItemPosition\0"
/* 10208 */ "topLevelEnterCallback\0"
/* 10230 */ "topLevelLeaveCallback\0"
/* 10252 */ "topOffset\0"
/* 10262 */ "topPosition\0"
/* 10274 */ "topShadowColor\0"
/* 10289 */ "topShadowPixmap\0"
/* 10305 */ "topWidget\0"
/* 10315 */ "transferProc\0"
/* 10328 */ "transferStatus\0"
/* 10343 */ "traversalCallback\0"
/* 10361 */ "traversalOn\0"
/* 10373 */ "traversalType\0"
/* 10387 */ "traverseObscuredCallback\0"
/* 10412 */ "treeUpdateProc\0"
/* 10427 */ "troughColor\0"
/* 10439 */ "unitType\0"
/* 10448 */ "unmapCallback\0"
/* 10462 */ "unpostBehavior\0"
/* 10477 */ "unselectPixmap\0"
/* 10492 */ "updateSliderSize\0"
/* 10509 */ "useAsyncGeometry\0"
/* 10526 */ "userData\0"
/* 10535 */ "validCursorForeground\0"
/* 10557 */ "valueChangedCallback\0"
/* 10578 */ "valueWcs\0"
/* 10587 */ "verifyBell\0"
/* 10598 */ "verticalFontUnit\0"
/* 10615 */ "verticalScrollBar\0"
/* 10633 */ "verticalSpacing\0"
/* 10649 */ "visibleItemCount\0"
/* 10666 */ "visibleWhenOff\0"
/* 10681 */ "visualPolicy\0"
/* 10694 */ "whichButton\0"
/* 10706 */ "wordWrap\0"
/* 10715 */ "workWindow\0"
/* 10726 */ "Alignment\0"
/* 10736 */ "AnimationMask\0"
/* 10750 */ "AnimationPixmap\0"
/* 10766 */ "AnimationStyle\0"
/* 10781 */ "ArrowDirection\0"
/* 10796 */ "AtomList\0"
/* 10805 */ "Attachment\0"
/* 10816 */ "AudibleWarning\0"
/* 10831 */ "Availability\0"
/* 10844 */ "BackgroundPixmap\0"
/* 10861 */ "BlendModel\0"
/* 10872 */ "BooleanDimension\0"
/* 10889 */ "BottomShadowPixmap\0"
/* 10908 */ "ButtonType\0"
/* 10919 */ "CallbackProc\0"
/* 10932 */ "Char\0"
/* 10937 */ "CharSetTable\0"
/* 10950 */ "ChildHorizontalAlignment\0"
/* 10975 */ "ChildPlacement\0"
/* 10990 */ "ChildType\0"
/* 11000 */ "ChildVerticalAlignment\0"
/* 11023 */ "CommandWindowLocation\0"
/* 11045 */ "CompoundText\0"
/* 11058 */ "DefaultButtonType\0"
/* 11076 */ "DeleteResponse\0"
/* 11091 */ "DialogStyle\0"
/* 11103 */ "DialogType\0"
/* 11114 */ "DoubleClickInterval\0"
/* 11134 */ "DragInitiatorProtocolStyle\0"
/* 11161 */ "DragReceiverProtocolStyle\0"
/* 11187 */ "DropSiteActivity\0"
/* 11204 */ "DropSiteOperations\0"
/* 11223 */ "DropSiteType\0"
/* 11236 */ "DropTransfers\0"
/* 11250 */ "ExtensionType\0"
/* 11264 */ "FileTypeMask\0"
/* 11277 */ "FontList\0"
/* 11286 */ "GadgetPixmap\0"
/* 11299 */ "HighlightPixmap\0"
/* 11315 */ "HorizontalDimension\0"
/* 11335 */ "HorizontalInt\0"
/* 11349 */ "HorizontalPosition\0"
/* 11368 */ "IconAttachment\0"
/* 11383 */ "ImportTargets\0"
/* 11397 */ "IndicatorType\0"
/* 11411 */ "ItemCount\0"
/* 11421 */ "Items\0"
/* 11427 */ "KeySym\0"
/* 11434 */ "KeySymTable\0"
/* 11446 */ "KeyboardFocusPolicy\0"
/* 11466 */ "LabelType\0"
/* 11476 */ "ListMarginHeight\0"
/* 11493 */ "ListMarginWidth\0"
/* 11509 */ "ListSizePolicy\0"
/* 11524 */ "ListSpacing\0"
/* 11536 */ "ManBottomShadowPixmap\0"
/* 11558 */ "ManForegroundPixmap\0"
/* 11578 */ "ManHighlightPixmap\0"
/* 11597 */ "ManTopShadowPixmap\0"
/* 11616 */ "MenuWidget\0"
/* 11627 */ "Mnemonic\0"
/* 11636 */ "MultiClick\0"
/* 11647 */ "NavigationType\0"
/* 11662 */ "Packing\0"
/* 11670 */ "PrimForegroundPixmap\0"
/* 11691 */ "Proc\0"
/* 11696 */ "ProcessingDirection\0"
/* 11716 */ "RectangleList\0"
/* 11730 */ "ResizePolicy\0"
/* 11743 */ "RowColumnType\0"
/* 11757 */ "ScrollBarDisplayPolicy\0"
/* 11780 */ "ScrollBarPlacement\0"
/* 11799 */ "ScrollingPolicy\0"
/* 11815 */ "SelectedItemCount\0"
/* 11833 */ "SelectedItems\0"
/* 11847 */ "SelectionPolicy\0"
/* 11863 */ "SelectionType\0"
/* 11877 */ "SeparatorType\0"
/* 11891 */ "ShadowType\0"
/* 11902 */ "ShellHorizDim\0"
/* 11916 */ "ShellHorizPos\0"
/* 11930 */ "ShellUnitType\0"
/* 11944 */ "ShellVertDim\0"
/* 11957 */ "ShellVertPos\0"
/* 11970 */ "SizePolicy\0"
/* 11981 */ "StringDirection\0"
/* 11997 */ "TearOffModel\0"
/* 12010 */ "TopShadowPixmap\0"
/* 12026 */ "TransferStatus\0"
/* 12041 */ "TraversalType\0"
/* 12055 */ "UnitType\0"
/* 12064 */ "UnpostBehavior\0"
/* 12079 */ "ValueWcs\0"
/* 12088 */ "VerticalAlignment\0"
/* 12106 */ "VerticalDimension\0"
/* 12124 */ "VerticalInt\0"
/* 12136 */ "VerticalPosition\0"
/* 12153 */ "VirtualBinding\0"
/* 12168 */ "VisibleItemCount\0"
/* 12185 */ "VisualPolicy\0"
/* 12198 */ "WhichButton\0"
/* 12210 */ "XmBackgroundPixmap\0"
/* 12229 */ "XmString\0"
/* 12238 */ "XmStringCharSet\0"
/* 12254 */ "XmStringTable\0"
/* 12268 */ "osfActivate\0"
/* 12280 */ "osfAddMode\0"
/* 12291 */ "osfBackSpace\0"
/* 12304 */ "osfBeginLine\0"
/* 12317 */ "osfCancel\0"
/* 12327 */ "osfClear\0"
/* 12336 */ "osfCopy\0"
/* 12344 */ "osfCut\0"
/* 12351 */ "osfDelete\0"
/* 12361 */ "osfDown\0"
/* 12369 */ "osfEndLine\0"
/* 12380 */ "osfHelp\0"
/* 12388 */ "osfInsert\0"
/* 12398 */ "osfLeft\0"
/* 12406 */ "osfMenu\0"
/* 12414 */ "osfMenuBar\0"
/* 12425 */ "osfPageDown\0"
/* 12437 */ "osfPageLeft\0"
/* 12449 */ "osfPageRight\0"
/* 12462 */ "osfPageUp\0"
/* 12472 */ "osfPaste\0"
/* 12481 */ "osfPrimaryPaste\0"
/* 12497 */ "osfQuickPaste\0"
/* 12511 */ "osfRight\0"
/* 12520 */ "osfSelect\0"
/* 12530 */ "osfUndo\0"
/* 12538 */ "osfUp\0"
/* 12544 */ "FONTLIST_DEFAULT_TAG_STRING\0"
/* 12572 */ "XmFONTLIST_DEFAULT_TAG_STRING\0"
/* 12602 */ "TopItemPosition\0"
/* 12618 */ "operatingCursorIcon\0"
/* ????? */ 
"\0";
